
    call DetectGameStarted()
endfunction



//***************************************************************************
//*
//*  Random distribution
//*
//*  Used to select a random object from a given distribution of chances
//*
//*  - RandomDistReset clears the distribution list
//*
//*  - RandomDistAddItem adds a new object to the distribution list
//*    with a given identifier and an integer chance to be chosen
//*
//*  - RandomDistChoose will use the current distribution list to choose
//*    one of the objects randomly based on the chance distribution
//*  
//*  Note that the chances are effectively normalized by their sum,
//*  so only the relative values of each chance are important
//*
//***************************************************************************

//===========================================================================
function RandomDistReset takes nothing returns nothing
    set bj_randDistCount = 0
endfunction

//===========================================================================
function RandomDistAddItem takes integer inID, integer inChance returns nothing
    set bj_randDistID[bj_randDistCount] = inID
    set bj_randDistChance[bj_randDistCount] = inChance
    set bj_randDistCount = bj_randDistCount + 1
endfunction

//===========================================================================
function RandomDistChoose takes nothing returns integer
    local integer sum = 0
	local integer chance = 0
	local integer index
	local integer foundID = -1
	local boolean done

	// No items?
	if (bj_randDistCount == 0) then
        return -1
    endif

	// Find sum of all chances
	set index = 0
	loop
        set sum = sum + bj_randDistChance[index]

		set index = index + 1
        exitwhen index == bj_randDistCount
    endloop

	// Choose random number within the total range
	set chance = GetRandomInt(1, sum)

	// Find ID which corresponds to this chance
	set index = 0
	set sum = 0
	set done = false
	loop
        set sum = sum + bj_randDistChance[index]

		if (chance <= sum) then
            set foundID = bj_randDistID[index]
			set done = true
		endif

		set index = index + 1
		if (index == bj_randDistCount) then
            set done = true
		endif

        exitwhen done == true
    endloop

	return foundID
endfunction



//***************************************************************************
//*
//*  Drop item
//*
//*  Makes the given unit drop the given item
//*
//*  Note: This could potentially cause problems if the unit is standing
//*        right on the edge of an unpathable area and happens to drop the
//*        item into the unpathable area where nobody can get it...
//*
//***************************************************************************

function UnitDropItem takes unit inUnit, integer inItemID returns item
    local real x
	local real y
	local real radius = 32
	local real unitX
	local real unitY
    local item droppedItem

	if (inItemID == -1) then
        return null
    endif

    set unitX = GetUnitX(inUnit)
	set unitY = GetUnitY(inUnit)

	set x = GetRandomReal(unitX - radius, unitX + radius)
	set y = GetRandomReal(unitY - radius, unitY + radius)

    set droppedItem = CreateItem(inItemID, x, y)

    call SetItemDropID(droppedItem, GetUnitTypeId(inUnit))
    call UpdateStockAvailability(droppedItem)

    return droppedItem
endfunction

//===========================================================================
function WidgetDropItem takes widget inWidget, integer inItemID returns item
    local real x
	local real y
	local real radius = 32
	local real widgetX
	local real widgetY

	if (inItemID == -1) then
        return null
    endif

    set widgetX = GetWidgetX(inWidget)
	set widgetY = GetWidgetY(inWidget)

	set x = GetRandomReal(widgetX - radius, widgetX + radius)
	set y = GetRandomReal(widgetY - radius, widgetY + radius)

    return CreateItem(inItemID, x, y)
endfunction
